@@ -1,5 +1,10 @@
+import enum
 import os
 import re
+import sys
+
+if sys.version_info < (3, 11):
+    enum.StrEnum = enum.Enum  # type: ignore[assignment]
 
 import pytest
 
@@ -247,15 +252,95 @@ def test_multiple_envvar(runner):
     assert result.output == "foo|bar\n"
 
 
-def test_trailing_blanks_boolean_envvar(runner):
+@pytest.mark.parametrize(
+    ("name", "value", "expected"),
+    (
+        # Lower-cased variations of value.
+        ("SHOUT", "true", True),
+        ("SHOUT", "false", False),
+        # Title-cased variations of value.
+        ("SHOUT", "True", True),
+        ("SHOUT", "False", False),
+        # Upper-cased variations of value.
+        ("SHOUT", "TRUE", True),
+        ("SHOUT", "FALSE", False),
+        # Random-cased variations of value.
+        ("SHOUT", "TruE", True),
+        ("SHOUT", "truE", True),
+        ("SHOUT", "FaLsE", False),
+        ("SHOUT", "falsE", False),
+        # Extra spaces around the value.
+        ("SHOUT", "true ", True),
+        ("SHOUT", " true", True),
+        ("SHOUT", " true ", True),
+        ("SHOUT", "false ", False),
+        ("SHOUT", " false", False),
+        ("SHOUT", " false ", False),
+        # Integer variations.
+        ("SHOUT", "1", True),
+        ("SHOUT", "0", False),
+        # Alternative states.
+        ("SHOUT", "t", True),
+        ("SHOUT", "f", False),
+        ("SHOUT", "y", True),
+        ("SHOUT", "n", False),
+        ("SHOUT", "yes", True),
+        ("SHOUT", "no", False),
+        ("SHOUT", "on", True),
+        ("SHOUT", "off", False),
+        # Blank value variations.
+        ("SHOUT", None, False),
+        ("SHOUT", "", False),
+        ("SHOUT", " ", False),
+        ("SHOUT", "       ", False),
+        # Variable names are not stripped of spaces and so don't match the
+        # flag, which then naturraly takes its default value.
+        ("SHOUT  ", "True", False),
+        ("SHOUT  ", "False", False),
+        ("  SHOUT  ", "True", False),
+        ("  SHOUT  ", "False", False),
+        # Same for random and reverse environment variable names: they are not
+        # recognized by the option.
+        ("RANDOM", "True", False),
+        ("NO_SHOUT", "True", False),
+        ("NO_SHOUT", "False", False),
+    ),
+)
+def test_boolean_envvar_normalization(runner, name, value, expected):
     @click.command()
     @click.option("--shout/--no-shout", envvar="SHOUT")
     def cli(shout):
         click.echo(f"shout: {shout!r}")
 
-    result = runner.invoke(cli, [], env={"SHOUT": " true "})
+    result = runner.invoke(cli, [], env={name: value})
     assert result.exit_code == 0
-    assert result.output == "shout: True\n"
+    assert result.output == f"shout: {expected}\n"
+
+
+@pytest.mark.parametrize(
+    ("name", "value"),
+    (
+        ("SHOUT", "tr ue"),
+        ("SHOUT", "10"),
+        ("SHOUT", "01"),
+        ("SHOUT", "00"),
+        ("SHOUT", "11"),
+        ("SHOUT", "randomstring"),
+        ("SHOUT", "None"),
+    ),
+)
+def test_boolean_envvar_bad_values(runner, name, value):
+    @click.command()
+    @click.option("--shout/--no-shout", envvar="SHOUT")
+    def cli(shout):
+        click.echo(f"shout: {shout!r}")
+
+    result = runner.invoke(cli, [], env={name: value})
+    assert result.exit_code == 2
+    assert (
+        f"Invalid value for '--shout': {value!r} is not a valid boolean."
+        in result.output
+    )
 
 
 def test_multiple_default_help(runner):
@@ -592,6 +677,13 @@ def test_missing_envvar(runner):
     assert result.exit_code == 2
     assert "Error: Missing option '--foo' (env var: 'bar')." in result.output
 
+    cli = click.Command(
+        "cli", params=[click.Option(["--foo"], show_envvar=True, required=True)]
+    )
+    result = runner.invoke(cli)
+    assert result.exit_code == 2
+    assert "Error: Missing option '--foo'." in result.output
+
 
 def test_case_insensitive_choice(runner):
     @click.command()
@@ -1007,6 +1099,172 @@ def test_type_from_flag_value():
     assert param.type is click.INT
 
 
+@pytest.mark.parametrize(
+    ("opt_params", "pass_flag", "expected"),
+    [
+        # Effect of the presence/absence of flag depending on type
+        ({"type": bool}, False, False),
+        ({"type": bool}, True, True),
+        ({"type": click.BOOL}, False, False),
+        ({"type": click.BOOL}, True, True),
+        ({"type": str}, False, "False"),
+        ({"type": str}, True, "True"),
+        # Effects of default value
+        ({"type": bool, "default": True}, False, True),
+        ({"type": bool, "default": True}, True, True),
+        ({"type": bool, "default": False}, False, False),
+        ({"type": bool, "default": False}, True, True),
+        ({"type": bool, "default": None}, False, False),
+        ({"type": bool, "default": None}, True, True),
+        ({"type": click.BOOL, "default": True}, False, True),
+        ({"type": click.BOOL, "default": True}, True, True),
+        ({"type": click.BOOL, "default": False}, False, False),
+        ({"type": click.BOOL, "default": False}, True, True),
+        ({"type": click.BOOL, "default": None}, False, False),
+        ({"type": click.BOOL, "default": None}, True, True),
+        ({"type": str, "default": True}, False, "True"),
+        ({"type": str, "default": True}, True, "True"),
+        ({"type": str, "default": False}, False, "False"),
+        ({"type": str, "default": False}, True, "True"),
+        ({"type": str, "default": "foo"}, False, "foo"),
+        ({"type": str, "default": "foo"}, True, "foo"),
+        ({"type": str, "default": None}, False, "False"),
+        ({"type": str, "default": None}, True, "True"),
+        # Effects of flag_value
+        ({"type": bool, "flag_value": True}, False, False),
+        ({"type": bool, "flag_value": True}, True, True),
+        ({"type": bool, "flag_value": False}, False, False),
+        ({"type": bool, "flag_value": False}, True, False),
+        ({"type": bool, "flag_value": None}, False, False),
+        ({"type": bool, "flag_value": None}, True, True),
+        ({"type": click.BOOL, "flag_value": True}, False, False),
+        ({"type": click.BOOL, "flag_value": True}, True, True),
+        ({"type": click.BOOL, "flag_value": False}, False, False),
+        ({"type": click.BOOL, "flag_value": False}, True, False),
+        ({"type": click.BOOL, "flag_value": None}, False, False),
+        ({"type": click.BOOL, "flag_value": None}, True, True),
+        ({"type": str, "flag_value": True}, False, "False"),
+        ({"type": str, "flag_value": True}, True, "True"),
+        ({"type": str, "flag_value": False}, False, "False"),
+        ({"type": str, "flag_value": False}, True, "False"),
+        ({"type": str, "flag_value": "foo"}, False, "False"),
+        ({"type": str, "flag_value": "foo"}, True, "foo"),
+        ({"type": str, "flag_value": None}, False, "False"),
+        ({"type": str, "flag_value": None}, True, "True"),
+        # Not passing --foo returns the default value as-is
+        ({"type": bool, "default": True, "flag_value": True}, False, True),
+        ({"type": bool, "default": True, "flag_value": False}, False, True),
+        ({"type": bool, "default": False, "flag_value": True}, False, False),
+        ({"type": bool, "default": False, "flag_value": False}, False, False),
+        ({"type": bool, "default": None, "flag_value": True}, False, False),
+        ({"type": bool, "default": None, "flag_value": False}, False, False),
+        ({"type": str, "default": True, "flag_value": True}, False, "True"),
+        ({"type": str, "default": True, "flag_value": False}, False, "True"),
+        ({"type": str, "default": False, "flag_value": True}, False, "False"),
+        ({"type": str, "default": False, "flag_value": False}, False, "False"),
+        ({"type": str, "default": "foo", "flag_value": True}, False, "foo"),
+        ({"type": str, "default": "foo", "flag_value": False}, False, "foo"),
+        ({"type": str, "default": "foo", "flag_value": "bar"}, False, "foo"),
+        ({"type": str, "default": "foo", "flag_value": None}, False, "foo"),
+        ({"type": str, "default": None, "flag_value": True}, False, "False"),
+        ({"type": str, "default": None, "flag_value": False}, False, "False"),
+        # Passing --foo returns the explicitly set flag_value
+        ({"type": bool, "default": True, "flag_value": True}, True, True),
+        ({"type": bool, "default": True, "flag_value": False}, True, False),
+        ({"type": bool, "default": False, "flag_value": True}, True, True),
+        ({"type": bool, "default": False, "flag_value": False}, True, False),
+        ({"type": bool, "default": None, "flag_value": True}, True, True),
+        ({"type": bool, "default": None, "flag_value": False}, True, False),
+        ({"type": str, "default": True, "flag_value": True}, True, "True"),
+        ({"type": str, "default": True, "flag_value": False}, True, "False"),
+        ({"type": str, "default": False, "flag_value": True}, True, "True"),
+        ({"type": str, "default": False, "flag_value": False}, True, "False"),
+        ({"type": str, "default": "foo", "flag_value": True}, True, "True"),
+        ({"type": str, "default": "foo", "flag_value": False}, True, "False"),
+        ({"type": str, "default": "foo", "flag_value": "bar"}, True, "bar"),
+        ({"type": str, "default": "foo", "flag_value": None}, True, "foo"),
+        ({"type": str, "default": None, "flag_value": True}, True, "True"),
+        ({"type": str, "default": None, "flag_value": False}, True, "False"),
+    ],
+)
+def test_flag_value_and_default(runner, opt_params, pass_flag, expected):
+    @click.command()
+    @click.option("--foo", is_flag=True, **opt_params)
+    def cmd(foo):
+        click.echo(repr(foo))
+
+    result = runner.invoke(cmd, ["--foo"] if pass_flag else [])
+    assert result.output == f"{expected!r}\n"
+
+
+@pytest.mark.parametrize(
+    "opts",
+    [
+        {"type": bool, "default": "foo"},
+        {"type": bool, "flag_value": "foo"},
+        {"type": click.BOOL, "default": "foo"},
+        {"type": click.BOOL, "flag_value": "foo"},
+    ],
+)
+def test_invalid_flag_definition(runner, opts):
+    @click.command()
+    @click.option("--foo", is_flag=True, **opts)
+    def cmd(foo):
+        click.echo(foo)
+
+    result = runner.invoke(cmd, ["--foo"])
+    assert (
+        "Error: Invalid value for '--foo': 'foo' is not a valid boolean"
+        in result.output
+    )
+
+
+@pytest.mark.parametrize(
+    ("flag_value", "envvar_value", "expected"),
+    [
+        # Envvar is set to false, so the flag default value is returned.
+        ("bar", "False", "False"),
+        # Same as above with alternative states and blank values.
+        ("bar", "0", "False"),
+        ("bar", "f", "False"),
+        ("bar", "", "False"),
+        # Envvar is set to true, so the flag_value is returned.
+        ("bar", "True", "bar"),
+        ("bar", "1", "bar"),
+        ("bar", "t", "bar"),
+        # So far we have the same cases as the test_flag_value_and_default
+        # test case. Now instead of passing a boolean-like value, let's use
+        # the flag_value as the envvar value.
+        ("bar", "bar", "bar"),
+        # flag_value is expected to be the exact same in the envvar for the flag to be
+        # activated.
+        ("bar", " bar ", "False"),
+        ("bar", "BAR", "False"),
+        ("bar", "random", "False"),
+        ("bar", "bar random", "False"),
+        ("bar", "random bar", "False"),
+        ("BAR", "BAR", "BAR"),
+        ("BAR", "bar", "False"),
+        (" bar ", " bar ", " bar "),
+        (" bar ", "bar", "False"),
+        (" bar ", "BAR", "False"),
+    ],
+)
+def test_envvar_flag_value(runner, flag_value, envvar_value, expected):
+    """Ensure that flag_value is recognized by the envvar."""
+
+    @click.command()
+    @click.option(
+        "--upper", type=str, is_flag=True, flag_value=flag_value, envvar="UPPER"
+    )
+    def cmd(upper):
+        click.echo(repr(upper))
+
+    result = runner.invoke(cmd, env={"UPPER": envvar_value})
+    assert result.exit_code == 0
+    assert result.output == f"{expected!r}\n"
+
+
 @pytest.mark.parametrize(
     ("option", "expected"),
     [
@@ -1022,7 +1280,7 @@ def test_type_from_flag_value():
         pytest.param(Option(["-a"], flag_value=True), True, id="bool flag_value"),
     ],
 )
-def test_is_bool_flag_is_correctly_set(option, expected):
+def test_bool_flag_auto_detection(option, expected):
     assert option.is_bool_flag is expected
 
 
@@ -1040,6 +1298,47 @@ def test_invalid_flag_combinations(runner, kwargs, message):
     assert message in str(e.value)
 
 
+def test_non_flag_with_non_negatable_default(runner):
+    class NonNegatable:
+        def __bool__(self):
+            raise ValueError("Cannot negate this object")
+
+    @click.command()
+    @click.option("--foo", default=NonNegatable())
+    def cmd(foo):
+        pass
+
+    result = runner.invoke(cmd)
+    assert result.exit_code == 0
+
+
+class HashType(enum.Enum):
+    MD5 = enum.auto()
+    SHA1 = enum.auto()
+
+
+class Number(enum.IntEnum):
+    ONE = enum.auto()
+    TWO = enum.auto()
+
+
+class Letter(enum.StrEnum):
+    A = enum.auto()
+    B = enum.auto()
+
+
+class Color(enum.Flag):
+    RED = enum.auto()
+    GREEN = enum.auto()
+    BLUE = enum.auto()
+
+
+class ColorInt(enum.IntFlag):
+    RED = enum.auto()
+    GREEN = enum.auto()
+    BLUE = enum.auto()
+
+
 @pytest.mark.parametrize(
     ("choices", "metavars"),
     [
@@ -1048,6 +1347,11 @@ def test_invalid_flag_combinations(runner, kwargs, message):
         pytest.param([1.0, 2.0], "[FLOAT]", id="float choices"),
         pytest.param([True, False], "[BOOLEAN]", id="bool choices"),
         pytest.param(["foo", 1], "[TEXT|INTEGER]", id="text/int choices"),
+        pytest.param(HashType, "[HASHTYPE]", id="enum choices"),
+        pytest.param(Number, "[NUMBER]", id="int enum choices"),
+        pytest.param(Letter, "[LETTER]", id="str enum choices"),
+        pytest.param(Color, "[COLOR]", id="flag enum choices"),
+        pytest.param(ColorInt, "[COLORINT]", id="int flag enum choices"),
     ],
 )
 def test_usage_show_choices(runner, choices, metavars):
@@ -1070,12 +1374,61 @@ def test_usage_show_choices(runner, choices, metavars):
         pass
 
     result = runner.invoke(cli_with_choices, ["--help"])
-    assert f"[{'|'.join([str(i) for i in choices])}]" in result.output
+    assert (
+        f"[{'|'.join(i.name if isinstance(i, enum.Enum) else str(i) for i in choices)}]"
+        in result.output
+    )
 
     result = runner.invoke(cli_without_choices, ["--help"])
     assert metavars in result.output
 
 
+@pytest.mark.parametrize(
+    ("choices", "default", "default_string"),
+    [
+        (["foo", "bar"], "bar", "bar"),
+        # The default value is not enforced to be in the choices.
+        (["foo", "bar"], "random", "random"),
+        # None cannot be a default value as-is: it left the default value as unset.
+        (["foo", "bar"], None, None),
+        ([0, 1], 0, "0"),
+        # Values are not coerced to the type of the choice, even if equivalent.
+        ([0, 1], 0.0, "0.0"),
+        ([1, 2], 2, "2"),
+        ([1.0, 2.0], 2, "2"),
+        ([1.0, 2.0], 2.0, "2.0"),
+        ([True, False], True, "True"),
+        ([True, False], False, "False"),
+        (["foo", 1], "foo", "foo"),
+        (["foo", 1], 1, "1"),
+        # Enum choices are rendered as their names.
+        # See: https://github.com/pallets/click/issues/2911
+        (HashType, HashType.SHA1, "SHA1"),
+        # Enum choices allow defaults strings that are their names.
+        (HashType, "SHA1", "SHA1"),
+        (Number, Number.TWO, "TWO"),
+        (Letter, Letter.B, "B"),
+        (Color, Color.GREEN, "GREEN"),
+        (ColorInt, ColorInt.GREEN, "GREEN"),
+    ],
+)
+def test_choice_default_rendering(runner, choices, default, default_string):
+    @click.command()
+    @click.option("-g", type=click.Choice(choices), default=default, show_default=True)
+    def cli_with_choices(g):
+        pass
+
+    # Check that the default value is kept normalized to the type of the choice.
+    assert cli_with_choices.params[0].default == default
+
+    result = runner.invoke(cli_with_choices, ["--help"])
+    extra_usage = f"[default: {default_string}]"
+    if default_string is None:
+        assert extra_usage not in result.output
+    else:
+        assert extra_usage in result.output
+
+
 @pytest.mark.parametrize(
     "opts_one,opts_two",
     [
@@ -1100,3 +1453,33 @@ def test_duplicate_names_warning(runner, opts_one, opts_two):
 
     with pytest.warns(UserWarning):
         runner.invoke(cli, [])
+
+
+@pytest.mark.parametrize(
+    ("multiple", "nargs", "default", "expected_message"),
+    [
+        (
+            False,  # multiple
+            2,  # nargs
+            42,  # default (not a list)
+            "'default' must be a list when 'nargs' != 1.",
+        ),
+        (
+            True,  # multiple
+            2,  # nargs
+            [
+                "test string which is not a list in the list"
+            ],  # default (list of non-lists)
+            "'default' must be a list of lists when 'multiple' is true"
+            " and 'nargs' != 1.",
+        ),
+    ],
+)
+def test_default_type_error_raises(multiple, nargs, default, expected_message):
+    with pytest.raises(ValueError, match=expected_message):
+        click.Option(
+            ["--test"],
+            multiple=multiple,
+            nargs=nargs,
+            default=default,
+        )
